%%Linear Quadratic Regulator
%an LQR is an implementation of control using the state space model. it
%attmepts to move the current state to some desired state. for conveniece
%the desired state is generally set to [0,0....], As maniplating the system
%to consider any other state from here is trivial (current-desired) -> 0
%the system from the kalman tutorial is considered here


clear all;
clc;
clf;

N = 500;
Ts = 0.02; %50Hz sampling
Time = 1:1:N;
Time = Time*Ts;

A = [0 1 0 0;
    0 0 0 0;
    0 0 0 1;
    0 0 0 0];

B = [0 1 0 0;
    0 0 0 1]'; %we have the ability to apply a force in the X axis and Y axis

C = [1 0 0 0;
    0 0 1 0]; %and are concerned with the objects position

D = 0;
sys = ss(A,B,C,D);

sysd = c2d(sys,Ts); %Discretise the continuous ss sys


input = [zeros(5,2);ones(N-5,2)]';


Q = [0 0 0 0;
    0 0.1 0 0;
    0 0 0 0;
    0 0 0 0.1]; %Covariance state

Q = Q + ones(4)*10^-10;
WQ = Q^(-1/2);
DWQ = inv(WQ);

noiseState = DWQ*randn(4,N);

R = [2 0;
    0 1]; %covariance sensors
R = R + ones(2)*10^-10;
WR = R^(-1/2); %take a variable with covar mat R to a whitened state
DWR = inv(WR); %dewhiten a white unit power GRV to covar R


noiseSensors = DWR*randn(2,N); %producing noise with the given covariance


%%Steady state KF
Lupanyov = zeros(4);
for i =1:1:1000
    Lupanyov = sysd.A*Lupanyov*sysd.A' + Q - sysd.A*Lupanyov*sysd.C'*inv(sysd.C*Lupanyov*sysd.C' + R)*sysd.C*Lupanyov*sysd.A';
end
kalmanGain = sysd.A*Lupanyov*sysd.C'*inv(sysd.C*Lupanyov*sysd.C' + R); %our state estimate filter



%%Steady state LQR gain
Q = zeros(4);
Q(1,1) = 1000;
Q(2,2) = 0.2;
Q(3,3) = 1000;
Q(4,4) = 0.3;

R = zeros(2);
R(1,1) = 0.2;
R(2,2) = 0.1;

LQRGain = lqrd(A,B,Q,R,Ts);


state = [20;10;-5;-15]; %True inital state 
estimatedState = [0;0;0;0]; %%inital state estimate

input = [0;0];
%%We cannot simulate the motion independent of the controller+kalman
%%filter, here it will all be considered as one.
for i = 2:1:N %%motion simulation
    
    state(:,i) = sysd.A*state(:,i-1) + sysd.B*input(:,i-1) + noiseState(:,i-1); %compute the true state position
    Measurement(:,i) = sysd.C*state(:,i-1) + noiseSensors(:,i-1);   %compute the observed position of the plant

    predictedState = sysd.A*estimatedState(:,i-1) + sysd.B*input(:,i-1); % predict the system position. note the control effort input is included in the state prediction    
    innovation = Measurement(:,i) - sysd.C*predictedState; 
    estimatedState(:,i) = predictedState + kalmanGain*innovation; %update system position accounting for the measured position
    
    
    input(:,i) = -LQRGain* estimatedState(:,i); %compute the optimal control effort input given the LQRGain. note the LQRgain is fixed, it does not depend on the current state
end

figure(1)
hold on
plot(Time,Measurement');
plot(Time,state([1,3],:)','k');
plot(Time,estimatedState([1,3],:)','b--');

figure(2)
plot(input','k');



